"""Mutation Test Result HTML Reporter."""

from __future__ import annotations

import datetime
import re
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from jinja2 import Environment, PackageLoader

from poodle import __version__ as poodle_version
from poodle.data_types import MutantTrial, PoodleConfig, TestingResults, TestingSummary

if TYPE_CHECKING:
    from collections.abc import Generator


def template_path() -> Path:
    """Return the path to the HTML Template folder."""
    return Path(__file__).parent.parent / "templates"


STATIC_FILES = [
    "html-report.js",
    "html-report.css",
    "poodle.ico",
]


def report_html(config: PoodleConfig, echo: Callable, testing_results: TestingResults, *_, **__) -> None:
    """Build HTML Report for Testing Results."""
    html_options = config.reporter_opts.get("html", {})
    if not isinstance(html_options, dict):
        raise TypeError("HTML Reporter Options (reporter_opts.html) must be a Dictionary.")

    report_folder = Path(html_options.get("report_folder", "mutation_reports"))
    copy_static_files(report_folder)

    env = Environment(loader=PackageLoader("poodle"), autoescape=True)

    common_vars = {
        "project": {
            "name": config.project_name,
            "version": config.project_version,
        },
        "poodle_version": poodle_version,
        "timestamp": local_timestamp(),
    }

    modules = module_data(testing_results, html_options)

    index_template = env.get_template("html-report-index.html.jinja")
    index_page = index_template.render(total=testing_results.summary, modules=modules, **common_vars)
    index_file = report_folder / "index.html"
    index_file.write_text(index_page.strip(), encoding="utf-8")

    module_template = env.get_template("html-report-module.html.jinja")
    for source_file, module in modules.items():
        module_page = module_template.render(source_file=source_file, module=module, **common_vars)
        (report_folder / module["report_file"]).write_text(module_page.strip(), encoding="utf-8")

    echo(f"HTML Report Generated at {index_file.resolve()}")


def copy_static_files(report_folder: Path) -> None:
    """Copy static files to report folder."""
    report_folder.mkdir(parents=True, exist_ok=True)
    templates_folder = template_path()
    for file in STATIC_FILES:
        shutil.copy2(templates_folder / file, report_folder / file)


def local_timestamp() -> str:
    """Return a local timestamp."""
    dt = datetime.datetime.now(datetime.timezone.utc)
    return dt.astimezone().strftime("%Y-%m-%d %H:%M:%S%z")


def module_data(testing_results: TestingResults, html_options: dict) -> dict[Path, dict[str, Any]]:
    """Return data for report pages."""
    modules = {trial.mutant.source_file for trial in testing_results.mutant_trials if trial.mutant.source_file}

    module_dict: dict[Path, dict] = {m: dict() for m in modules}  # noqa: C408 - `dict()` call is needed here
    include_found_index = html_options.get("include_found_trials_on_index", False)
    include_found_source = html_options.get("include_found_trials_with_source", True)

    for idx, module in enumerate(modules, start=1):
        module_dict[module]["report_file"] = f"module-{idx}.html"
        module_dict[module]["file_id"] = re.sub(r"[^A-Za-z0-9\-_:.]", "_", str(module))

        module_dict[module]["trials"] = list(module_trials(module, testing_results.mutant_trials))
        module_dict[module]["trials"].sort(key=lambda trial: trial.mutant.lineno)

        module_dict[module]["lines"] = list(module_lines(module))
        module_add_trials_to_lines(module_dict[module]["trials"], module_dict[module]["lines"], include_found_source)

        module_dict[module]["summary"] = module_summary(module_dict[module]["trials"])

        if not include_found_index:
            module_dict[module]["trials"] = remove_found_trials(module_dict[module]["trials"])

    return module_dict


def module_trials(
    module: Path,
    mutant_trials: list[MutantTrial],
) -> Generator[MutantTrial, Any, None]:
    """Return trials for a module."""
    for trial in mutant_trials:
        if trial.mutant.source_file == module:
            yield trial


def module_lines(module: Path) -> Generator[dict[str, Any], Any, None]:
    """Return lines for a module."""
    for lineno, line in enumerate(module.read_text("utf-8").splitlines(), start=1):
        yield {
            "lineno": lineno,
            "text": line,
            "trials": [],
            "row_class": "plain",
        }


def module_add_trials_to_lines(
    module_trials: list[MutantTrial],
    module_lines: list[dict[str, Any]],
    include_found: bool,
) -> None:
    """Add trials to lines."""
    for trial in module_trials:
        if include_found or not trial.result.found:
            line = module_lines[trial.mutant.lineno - 1]
            line["trials"].append(trial)
            if line["row_class"] == "plain":
                line["row_class"] = "found" if trial.result.found else "not_found"
            elif (line["row_class"] == "found" and not trial.result.found) or (
                line["row_class"] == "not_found" and trial.result.found
            ):
                line["row_class"] = "partial_found"


def module_summary(module_trials: list[MutantTrial]) -> TestingSummary:
    """Return summary for a module."""
    summary = TestingSummary()
    for trial in module_trials:
        summary += trial.result
    summary.trials = len(module_trials)
    return summary


def remove_found_trials(trials: list[MutantTrial]) -> list[MutantTrial]:
    """Remove not found trials from a list of trials."""
    return [trial for trial in trials if not trial.result.found]



"""Basic Mutation Test Result Reporters."""

from __future__ import annotations

from pathlib import Path
from typing import Callable

from poodle.data_types import PoodleConfig, TestingResults
from poodle.util import to_json


def get_include_statuses(config: PoodleConfig, prefix: str) -> set[bool]:
    """Get set of statuses to include in report."""
    include_statuses = set()
    if config.reporter_opts.get(f"{prefix}_report_found", False):
        include_statuses.add(True)  # noqa: FBT003
    if config.reporter_opts.get(f"{prefix}_report_not_found", True):
        include_statuses.add(False)  # noqa: FBT003
    return include_statuses


def report_summary(echo: Callable, testing_results: TestingResults, *_, **__) -> None:
    """Echo quick summary to console."""
    echo("")
    summary = testing_results.summary
    if summary.trials < 1:
        echo("!!! No mutants found to test !!!", fg="yellow")
        return

    echo("*** Results Summary ***", fg="green")
    echo(f"Testing found {summary.success_rate:.1%} of Mutants.")
    if summary.not_found:
        echo(f" - {summary.not_found} mutant(s) were not found.")
    if summary.timeout:
        echo(f" - {summary.timeout} mutant(s) caused trial to timeout.")
    if summary.errors:
        echo(f" - {summary.errors} mutant(s) could not be tested due to an error.")


def report_not_found(config: PoodleConfig, echo: Callable, testing_results: TestingResults, *_, **__) -> None:
    """Echo information about Trials that did not pass."""
    failed_trials = [trial for trial in testing_results.mutant_trials if not trial.result.found]
    if not failed_trials:
        return

    failed_trials.sort(
        key=lambda trial: (
            trial.mutant.source_folder,
            str(trial.mutant.source_file) or "",
            trial.mutant.lineno,
            trial.mutant.mutator_name,
        )
    )

    not_found_file = config.reporter_opts.get("not_found_file")

    echo("", file=not_found_file)
    echo("*** Mutants Not Found ***", fg="yellow", file=not_found_file)
    for trial in failed_trials:
        mutant = trial.mutant
        result = trial.result

        echo("", file=not_found_file)
        echo(
            f"Mutant Trial Result: {result.reason_code}",
            file=not_found_file,
        )
        echo(f"Mutator: {mutant.mutator_name}", file=not_found_file)
        if result.reason_desc:
            echo(result.reason_desc, file=not_found_file)

        if mutant.unified_diff:
            echo(mutant.unified_diff, file=not_found_file)
        else:
            echo(
                f"source_file={mutant.source_file} lineno={mutant.lineno} col_offset={mutant.col_offset} "
                f"end_lineno={mutant.end_lineno} end_col_offset={mutant.end_col_offset}",
                file=not_found_file,
            )
            echo("text:", file=not_found_file)
            echo(mutant.text, file=not_found_file)


def report_json(config: PoodleConfig, echo: Callable, testing_results: TestingResults, *_, **__) -> None:
    """Create JSON file with test results."""
    include_statuses = get_include_statuses(config, "json")
    mutant_trials = [trial for trial in testing_results.mutant_trials if trial.result.found in include_statuses]

    out_results = TestingResults(
        summary=testing_results.summary  # type: ignore [arg-type]
        if config.reporter_opts.get("json_include_summary", True)
        else None,
        mutant_trials=mutant_trials,
    )

    json_file = config.reporter_opts.get("json_report_file", "mutation-testing-report.json")
    if json_file == "sysout":
        echo(to_json(out_results, indent=4))
    else:
        Path(json_file).write_text(to_json(out_results))

    echo(f"JSON report written to {json_file!s}", fg="green")



"""Resolve configuration options and build PoodleConfig object."""

from __future__ import annotations

import logging
import os
from collections.abc import Iterable
from pathlib import Path
from typing import Any

from mergedeep import merge  # type: ignore[import-untyped]
from wcmatch import glob

from . import PoodleInputError, poodle_config, tomllib
from .data_types import PoodleConfig

default_source_folders = [Path("src"), Path("lib")]

default_log_format = "%(levelname)s [%(process)d] %(name)s.%(funcName)s:%(lineno)d - %(message)s"
default_log_level = logging.WARN

default_file_flags = glob.GLOBSTAR | glob.NODIR
default_file_filters = ["test_*.py", "*_test.py", "poodle_config.py", "setup.py"]
default_file_copy_flags = glob.GLOBSTAR | glob.NODIR
default_file_copy_filters = ["__pycache__/**"]
default_work_folder = Path(".poodle-temp")

default_min_timeout = 10
default_timeout_multiplier = 10
default_runner = "command_line"

default_reporters = ["summary", "not_found"]


def default_max_workers() -> int:
    """Calculate Default for max_workers as one less than available processors."""
    if hasattr(os, "sched_getaffinity"):
        return len(os.sched_getaffinity(0)) - 1
    cpu_count = os.cpu_count() or 1  # nomut: Number
    if cpu_count > 1:
        return cpu_count - 1
    return cpu_count


def build_config(  # noqa: PLR0913
    cmd_sources: tuple[Path],
    cmd_config_file: Path | None,
    cmd_quiet: int,
    cmd_verbose: int,
    cmd_max_workers: int | None,
    cmd_excludes: tuple[str],
    cmd_only_files: tuple[str],
    cmd_report: tuple[str],
    cmd_html: Path | None,
    cmd_json: Path | None,
    cmd_fail_under: float | None,
) -> PoodleConfig:
    """Build PoodleConfig object."""
    config_file_path = get_config_file_path(cmd_config_file)
    config_file_data = get_config_file_data(config_file_path)
    project_name, project_version = get_project_info(config_file_path)

    log_format = get_str_from_config("log_format", config_file_data, default=default_log_format)
    log_level: int | str = get_any_from_config(
        "log_level",
        config_file_data,
        default=default_log_level,
        command_line=get_cmd_line_log_level(cmd_quiet, cmd_verbose),
    )
    logging.basicConfig(format=log_format, level=log_level)

    file_filters = get_str_list_from_config("file_filters", config_file_data, default=default_file_filters)
    # TODO: append file excludes and append py excludes
    # file_filters += get_str_list_from_config("exclude", config_file_data, default=[]) # noqa: ERA001
    file_filters += cmd_excludes

    cmd_reporter_opts: dict[str, Any] = {}
    if cmd_html:
        merge(cmd_reporter_opts, {"html": {"report_folder": cmd_html}})
    if cmd_json:
        merge(cmd_reporter_opts, {"json_report_file": cmd_json})

    return PoodleConfig(
        project_name=get_str_from_config("project_name", config_file_data, default=project_name),
        project_version=get_str_from_config("project_version", config_file_data, default=project_version),
        config_file=config_file_path,
        source_folders=get_source_folders(cmd_sources, config_file_data),
        only_files=get_str_list_from_config("only_files", config_file_data, default=[], command_line=cmd_only_files),
        file_flags=get_int_from_config("file_flags", config_file_data, default=default_file_flags),
        file_filters=file_filters,
        file_copy_flags=get_int_from_config("file_copy_flags", config_file_data, default=default_file_copy_flags),
        file_copy_filters=get_str_list_from_config(
            "file_copy_filters",
            config_file_data,
            default=default_file_copy_filters,
        ),
        work_folder=get_path_from_config("work_folder", config_file_data, default=default_work_folder),
        max_workers=get_int_from_config(
            "max_workers",
            config_file_data,
            default=default_max_workers(),
            command_line=cmd_max_workers,
        ),
        log_format=log_format,
        log_level=log_level,
        echo_enabled=get_bool_from_config(
            "echo_enabled",
            config_file_data,
            default=True,
            command_line=get_cmd_line_echo_enabled(cmd_quiet),
        ),
        echo_no_color=get_bool_from_config("echo_no_color", config_file_data),
        mutator_opts=get_dict_from_config("mutator_opts", config_file_data),
        skip_mutators=get_str_list_from_config("skip_mutators", config_file_data, default=[]),
        add_mutators=get_any_list_from_config("add_mutators", config_file_data),
        min_timeout=get_int_from_config("min_timeout", config_file_data) or default_min_timeout,
        timeout_multiplier=get_int_from_config("timeout_multiplier", config_file_data) or default_timeout_multiplier,
        runner=get_str_from_config("runner", config_file_data, default=default_runner),
        runner_opts=get_dict_from_config("runner_opts", config_file_data),
        reporters=get_reporters(config_file_data, cmd_report, cmd_html, cmd_json),
        reporter_opts=get_dict_from_config("reporter_opts", config_file_data, command_line=cmd_reporter_opts),
        fail_under=get_float_from_config("fail_under", config_file_data, command_line=cmd_fail_under),
    )


def get_reporters(
    config_file_data: dict,
    cmd_report: tuple[str],
    cmd_html: Path | None,
    cmd_json: Path | None,
) -> list[str]:
    """Retrieve list of reporters to use."""
    reporters = get_str_list_from_config("reporters", config_file_data, default=default_reporters)
    reporters += [reporter for reporter in cmd_report if reporter not in reporters]
    if cmd_html:
        reporters.append("html")
    if cmd_json:
        reporters.append("json")
    return reporters


def get_cmd_line_log_level(cmd_quiet: int, cmd_verbose: int) -> int | None:
    """Map verbosity input to logging level."""
    if cmd_quiet >= 3:
        return logging.CRITICAL
    if cmd_quiet == 2:
        return logging.ERROR
    if cmd_quiet == 1:
        return logging.WARN

    if cmd_verbose >= 3:
        return logging.NOTSET
    if cmd_verbose == 2:
        return logging.DEBUG
    if cmd_verbose == 1:
        return logging.INFO

    return None


def get_cmd_line_echo_enabled(cmd_quiet: int) -> bool | None:
    """Map verbosity input to enable/disable echo statements."""
    if cmd_quiet == 0:
        return None
    return False


def get_config_file_path(config_file: Path | None) -> Path | None:
    """Identify which configuration file to use.

    Checks in this order, first value is used.
    Command Line, poodle.toml, pyproject.toml.
    Returns None if no config file found.
    """
    if config_file:
        if not config_file.is_file():
            msg = f"Config file not found: '{config_file}'"
            raise PoodleInputError(msg)
        return config_file

    if hasattr(poodle_config, "config_file"):
        config_file = Path(poodle_config.config_file)
        if not config_file.is_file():
            msg = f"config_file not found: '{poodle_config.config_file}'"
            raise PoodleInputError(msg)
        return config_file

    files = [
        "poodle.toml",
        "pyproject.toml",
    ]  # TODO: ["poodle.toml", "tox.ini", "setup.cfg", "pyproject.toml"]

    for file in files:
        path = Path(file)
        if path.is_file():
            return path

    return None


def get_config_file_data(config_file: Path | None) -> dict:
    """Retrieve Poodle configuration from specified Config File."""
    if not config_file:
        return {}

    if config_file.suffix == ".toml":
        return get_config_file_data_toml(config_file)

    # TODO: tox.ini and setup.cfg
    # https://tox.wiki/en/3.24.5/config.html

    msg = f"Config file type not supported: --config_file='{config_file}'"
    raise PoodleInputError(msg)


def get_project_info(config_file: Path | None) -> tuple[str, str]:
    """Retrieve Poodle configuration from specified Config File."""
    if not config_file:
        return ("", "")

    if config_file.suffix == ".toml":
        return get_project_info_toml(config_file)

    # TODO: tox.ini and setup.cfg
    # https://tox.wiki/en/3.24.5/config.html

    msg = f"Config file type not supported: --config_file='{config_file}'"
    raise PoodleInputError(msg)


def get_config_file_data_toml(config_file: Path) -> dict:
    """Retrieve Poodle configuration from a 'toml' Config File."""
    config_data = tomllib.load(config_file.open(mode="rb"))
    config_data: dict = config_data.get("tool", config_data)  # type: ignore[no-redef]
    return config_data.get("poodle", {})


def get_project_info_toml(config_file: Path) -> tuple[str, str]:
    """Retrieve Project Name and Version from a 'toml' Config File."""
    config_data = tomllib.load(config_file.open(mode="rb"))
    config_data: dict = config_data.get("project", config_data)  # type: ignore[no-redef]
    return config_data.get("name", ""), config_data.get("version", "")


def get_source_folders(command_line_sources: tuple[Path], config_data: dict) -> list[Path]:
    """Retrieve list of source folders that contain files to mutate.

    Verifies that all returned values are existing directories.
    """
    source_folders = get_path_list_from_config(
        option_name="source_folders",
        config_data=config_data,
        command_line=command_line_sources,
        default=[source for source in default_source_folders if source.is_dir()],
    )

    if not source_folders:
        raise PoodleInputError("No source folder found to mutate.")

    for source in source_folders:
        if not source.is_dir():
            msg = f"Source '{source}' must be a folder."
            raise PoodleInputError(msg)

    return source_folders


def get_bool_from_config(
    option_name: str,
    config_data: dict,
    default: bool | None = None,
    command_line: bool | str | None = None,
) -> bool | None:
    """Retrieve Config Option that should be a Boolean.

    Retrieve highest priority value from config sources.
    """
    value, _ = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if isinstance(value, bool):
        return value

    if isinstance(value, str):
        return value.upper() == "TRUE"

    return default


def get_path_from_config(
    option_name: str,
    config_data: dict,
    default: Path,
    command_line: Path | None = None,
) -> Path:
    """Retrieve Config Option that should be a String.

    Retrieve highest priority value from config sources.
    """
    value, source = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if value is None:
        return default

    try:
        return Path(value)
    except TypeError:
        msg = f"{option_name} from {source} must be a valid StrPath"
        raise PoodleInputError(msg) from None


def get_path_list_from_config(
    option_name: str,
    config_data: dict,
    default: list[Path] | None = None,
    command_line: tuple[Path] | None = None,
) -> list[Path]:
    """Retrieve Config Option that should be a List of Paths.

    Retrieve highest priority value from config sources.
    If input was a single Path, return as a list of Paths.
    Convert input Iterable to List.
    """
    default_fix = default or []
    command_line_fix = command_line or ()

    values, source = get_option_from_config(
        option_name=option_name,
        config_data=config_data,
        command_line=command_line_fix,
    )

    if values is None:
        return default_fix

    try:
        if isinstance(values, Path):
            return [values]

        if isinstance(values, str):
            return [Path(values)]

        return [Path(value) for value in values]
    except TypeError:
        msg = f"{option_name} from {source} must be a valid Iterable[StrPath]"
        raise PoodleInputError(msg) from None


def get_any_from_config(
    option_name: str,
    config_data: dict,
    default: Any = None,  # noqa: ANN401
    command_line: Any | None = None,  # noqa: ANN401
) -> Any:  # noqa: ANN401
    """Retrieve Config Option that can by any type.

    Retrieve highest priority value from config sources.
    """
    value, _ = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if value is None:
        return default
    return value


def get_any_list_from_config(
    option_name: str,
    config_data: dict,
    default: list[Any] | None = None,
    command_line: tuple[Any] | None = None,
) -> list[Any]:
    """Retrieve Config Option that should be a List of any types.

    Retrieve highest priority value from config sources.
    Convert input Iterable to List.
    """
    default_fix = default or []
    command_line_fix = command_line or ()

    values, _ = get_option_from_config(
        option_name=option_name,
        config_data=config_data,
        command_line=command_line_fix,
    )

    if values is None:
        return default_fix

    if isinstance(values, str):
        return [values]

    if isinstance(values, Iterable):
        return list(values)

    return [values]


def get_int_from_config(
    option_name: str,
    config_data: dict,
    default: int | None = None,
    command_line: int | None = None,
) -> int | None:
    """Retrieve Config Option that should be an int or None.

    Retrieve highest priority value from config sources.
    """
    value, source = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if value is None:
        return default

    try:
        return int(value)
    except ValueError:
        msg = f"{option_name} from {source} must be a valid int"
        raise PoodleInputError(msg) from None


def get_float_from_config(
    option_name: str,
    config_data: dict,
    default: float | None = None,
    command_line: float | None = None,
) -> float | None:
    """Retrieve Config Option that should be an float or None.

    Retrieve highest priority value from config sources.
    """
    value, source = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if value is None:
        return default

    try:
        return float(value)
    except ValueError:
        msg = f"{option_name} from {source} must be a valid float"
        raise PoodleInputError(msg) from None


def get_str_from_config(
    option_name: str,
    config_data: dict,
    default: str = "",
    command_line: str = "",
) -> str:
    """Retrieve Config Option that should be a String.

    Retrieve highest priority value from config sources.
    """
    value, _ = get_option_from_config(option_name=option_name, config_data=config_data, command_line=command_line)

    if value is None:
        return default

    return str(value)


def get_str_list_from_config(
    option_name: str,
    config_data: dict,
    default: list[str] | None = None,
    command_line: str | tuple[str] | None = None,
) -> list[str]:
    """Retrieve Config Option that should be a List of Strings.

    Retrieve highest priority value from config sources.
    If input was a single string, return as a list of strings.
    Convert input Iterable to List.
    """
    default_fix = default or []
    command_line_fix = command_line or ()

    values, source = get_option_from_config(
        option_name=option_name,
        config_data=config_data,
        command_line=command_line_fix,
    )

    if values is None:
        return default_fix

    if isinstance(values, str):
        return [values]

    try:
        return [str(value) for value in values]
    except TypeError:
        msg = f"{option_name} from {source} must be a valid Iterable[str]"
        raise PoodleInputError(msg) from None


def get_option_from_config(
    option_name: str,
    config_data: dict,
    command_line: Any,  # noqa: ANN401
) -> tuple[Any | None, str | None]:
    """Retrieve Config Option of any type.

    Check sources in priority order, and return the first one found.

    Priority Order: Command Line, poodle_config.py, config file

    Returns: Identified Config value, Source Name
    """
    if command_line or command_line is False:
        return command_line, "Command Line"

    if hasattr(poodle_config, option_name):
        return getattr(poodle_config, option_name), "poodle_config.py"

    if option_name in config_data:
        return config_data[option_name], "config file"

    return None, None


def get_dict_from_config(
    option_name: str,
    config_data: dict,
    default: dict | None = None,
    command_line: dict | None = None,
) -> dict:
    """Merge Key-Value pairs from Config sources to a dict.

    Builds dict by starting with values from lowest priority source.
    Then 'updating' them with values from higher priority sources.

    Priority Order: Command Line, poodle_config.py, config file, defaults
    """
    default = default or {}
    command_line = command_line or {}

    option_value = default

    if option_name in config_data:
        try:
            merge(option_value, config_data[option_name])
        except TypeError:
            msg = f"{option_name} in config file must be a valid dict"
            raise PoodleInputError(msg) from None

    if hasattr(poodle_config, option_name):
        try:
            merge(option_value, getattr(poodle_config, option_name))
        except TypeError:
            msg = f"poodle_config.{option_name} must be a valid dict"
            raise PoodleInputError(msg) from None

    if command_line:
        merge(option_value, command_line)

    return option_value



from unittest import mock

import pytest

from poodle import report
from poodle.reporters import report_not_found, report_summary
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def mock_logger():
    with mock.patch("poodle.report.logger") as logger:
        yield logger


def test_logger():
    assert report.logger.name == "poodle.report"


class TestReporters:
    def test_builtin_reporters(self):
        assert report.builtin_reporters == {
            "summary": report_summary,
            "not_found": report_not_found,
            "json": report.report_json,
            "html": report.report_html,
        }

    def test_generate_reporters(self, mock_logger: mock.MagicMock):
        config = PoodleConfigStub(
            reporters=["summary", "not_found", "poodle.reporters.report_summary"],
        )
        gen_report = report.generate_reporters(config)

        assert next(gen_report) is report_summary
        mock_logger.debug.assert_called_with("Reporters: %s", config.reporters)

        assert next(gen_report) is report_not_found
        assert next(gen_report) is report_summary
        with pytest.raises(StopIteration):
            next(gen_report)



"""Poodle Mutation Test Tool."""

from __future__ import annotations

import importlib
import sys
from contextlib import suppress
from pathlib import Path
from typing import Any

__version__ = "1.3.0"


class PoodleTestingFailedError(Exception):
    """Poodle testing failed."""


class PoodleNoMutantsFoundError(Exception):
    """Poodle could not find any mutants to test."""


class PoodleInputError(ValueError):
    """An input value from Command Line, poodle_config.py, or a config file was invalid."""


class PoodleTrialRunError(Exception):
    """An unexpected error occurred when running a mutation trial or clean run."""


try:
    import tomllib  # type: ignore [import-not-found]
except ModuleNotFoundError:  # < py3.11
    import tomli as tomllib  # type: ignore [no-redef]


poodle_config: Any = None
with suppress(ImportError):
    sys.path.append(str(Path.cwd()))
    poodle_config = importlib.import_module("poodle_config")



"""Poodle core run process."""

from __future__ import annotations

import logging
import shutil

from . import PoodleNoMutantsFoundError, PoodleTestingFailedError, __version__
from .data_types import PoodleConfig, PoodleWork
from .mutate import create_mutants_for_all_mutators, initialize_mutators
from .report import generate_reporters
from .run import clean_run_each_source_folder, get_runner, run_mutant_trails
from .util import calc_timeout, create_temp_zips, create_unified_diff, pprint_str

logger = logging.getLogger(__name__)


def main_process(config: PoodleConfig) -> None:
    """Poodle core run process."""
    work = PoodleWork(config)  # sets logging defaults
    print_header(work)
    logger.info("\n%s", pprint_str(config))

    if config.work_folder.exists():
        logger.info("delete %s", config.work_folder)
        shutil.rmtree(config.work_folder)

    create_temp_zips(work)

    work.mutators = initialize_mutators(work)
    work.runner = get_runner(config)
    work.reporters = list(generate_reporters(config))

    mutants = create_mutants_for_all_mutators(work)
    if not mutants:
        raise PoodleNoMutantsFoundError("No mutants were found to test!")
    work.echo(f"Identified {len(mutants)} mutants")

    clean_run_results = clean_run_each_source_folder(work)
    timeout = calc_timeout(config, clean_run_results)
    results = run_mutant_trails(work, mutants, timeout)

    for trial in results.mutant_trials:
        trial.mutant.unified_diff = create_unified_diff(trial.mutant)

    for reporter in work.reporters:
        reporter(config=config, echo=work.echo, testing_results=results)

    logger.info("delete %s", config.work_folder)
    shutil.rmtree(config.work_folder)

    if config.fail_under and results.summary.success_rate < config.fail_under / 100:
        msg = f"Mutation score {results.summary.coverage_display} is below {config.fail_under:.2f}%"
        raise PoodleTestingFailedError(msg)


poodle_header_str = r"""
|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|\/|
    ____                  ____         ''',
   / __ \____  ____  ____/ / /__    o_)O \)____)"
  / /_/ / __ \/ __ \/ __  / / _ \    \_        )
 / ____/ /_/ / /_/ / /_/ / /  __/      '',,,,,,
/_/    \____/\____/\__,_/_/\___/         ||  ||
Mutation Tester Version {version:<15} "--'"--'
|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|/\|

"""


def print_header(work: PoodleWork) -> None:
    """Print a header to the console."""
    work.echo(poodle_header_str.format(version=__version__), fg="blue")
    work.echo("Running with the following configuration:")
    work.echo(f" - Source Folders: {[str(folder) for folder in work.config.source_folders]}")
    work.echo(f" - Config File:    {work.config.config_file}")
    work.echo(f" - Max Workers:    {work.config.max_workers}")
    work.echo(f" - Runner:         {work.config.runner}")
    work.echo(f" - Reporters:      {work.config.reporters}")
    work.echo()



"""Mutate Operators."""

from __future__ import annotations

import ast
from typing import Callable, ClassVar

from poodle.data_types import FileMutation, Mutator, PoodleConfig


class OperationMutator(ast.NodeVisitor, Mutator):
    """Base class for mutating operations."""

    # Binary Operators as of Python 3.12:
    # https://docs.python.org/3/library/ast.html#ast.BinOp
    # https://www.w3schools.com/python/python_operators.asp
    # ast.Add       +
    # ast.Sub       -
    # ast.Mult      *
    # ast.Div       /
    # ast.FloorDiv  //
    # ast.Mod       %
    # ast.Pow       **
    # ast.LShift    <<
    # ast.RShift    >>
    # ast.BitOr     |
    # ast.BitXor    ^
    # ast.BitAnd    &
    # ast.MatMult   @

    type_map_levels: ClassVar[dict[str, dict[type, list[type]]]] = {
        "min": {
            ast.Add: [ast.Mult],
            ast.Sub: [ast.Div],
            ast.Mult: [ast.Add],
            ast.Div: [ast.Sub],
            ast.FloorDiv: [ast.Div],
            ast.Mod: [ast.Sub],
            ast.Pow: [ast.Mult],
            ast.LShift: [ast.RShift],
            ast.RShift: [ast.LShift],
            ast.BitOr: [ast.BitAnd],
            ast.BitXor: [ast.BitOr],
            ast.BitAnd: [ast.BitXor],
        },
        "std": {
            ast.Add: [ast.Sub, ast.Mult],
            ast.Sub: [ast.Add, ast.Div],
            ast.Mult: [ast.Div, ast.Add],
            ast.Div: [ast.Mult, ast.Sub],
            ast.FloorDiv: [ast.Mult, ast.Div],
            ast.Mod: [ast.FloorDiv, ast.Sub],
            ast.Pow: [ast.Mult, ast.Div],
            ast.LShift: [ast.RShift],
            ast.RShift: [ast.LShift],
            ast.BitOr: [ast.BitAnd],
            ast.BitXor: [ast.BitOr, ast.BitAnd],
            ast.BitAnd: [ast.BitOr],
        },
        "max": {
            ast.Add: [ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod, ast.Pow],
            ast.Sub: [ast.Add, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod, ast.Pow],
            ast.Mult: [ast.Add, ast.Sub, ast.Div, ast.FloorDiv, ast.Mod, ast.Pow],
            ast.Div: [ast.Add, ast.Sub, ast.Mult, ast.FloorDiv, ast.Mod, ast.Pow],
            ast.FloorDiv: [ast.Add, ast.Sub, ast.Mult, ast.Div, ast.Mod, ast.Pow],
            ast.Mod: [ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Pow],
            ast.Pow: [ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod],
            ast.LShift: [ast.RShift, ast.BitOr, ast.BitXor, ast.BitAnd],
            ast.RShift: [ast.LShift, ast.BitOr, ast.BitXor, ast.BitAnd],
            ast.BitOr: [ast.LShift, ast.RShift, ast.BitXor, ast.BitAnd],
            ast.BitXor: [ast.LShift, ast.RShift, ast.BitOr, ast.BitAnd],
            ast.BitAnd: [ast.LShift, ast.RShift, ast.BitOr, ast.BitXor],
        },
    }

    def __init__(self, config: PoodleConfig, echo: Callable, *args, **kwargs) -> None:
        """Initialize and read settings."""
        super().__init__(config, echo, *args, **kwargs)
        self.mutants: list[FileMutation] = []

        level = self.config.mutator_opts.get("operator_level", "std")
        if level not in self.type_map_levels:
            echo(f"WARN: Invalid value operator_opts.operator_level={level}.  Using Default value 'std'")
            level = "std"

        self.type_map: dict[type, list[type]] = self.type_map_levels[level]

    def create_mutations(self, parsed_ast: ast.Module, *_, **__) -> list[FileMutation]:
        """Visit ast nodes and return created mutants."""
        self.mutants = []
        self.add_parent_attr(parsed_ast)
        self.visit(parsed_ast)
        return self.mutants


class BinaryOperationMutator(OperationMutator):
    """Mutate Binary Operations."""

    mutator_name = "BinOp"

    def visit_BinOp(self, node: ast.BinOp) -> None:
        """Identify replacement Operations and create Mutants."""
        if self.is_annotation(node):
            return

        if type(node.op) in self.type_map:
            mut_types = self.type_map[type(node.op)]
            self.mutants.extend([self.create_bin_op_mutant(node, new_type) for new_type in mut_types])

    def create_bin_op_mutant(self, node: ast.BinOp, new_type: type) -> FileMutation:
        """Create BinOp with replacement operation."""
        return self.create_file_mutation(node, ast.unparse(ast.BinOp(left=node.left, op=new_type(), right=node.right)))


class AugAssignMutator(OperationMutator):
    """Mutate Augmented Assignments."""

    mutator_name = "AugAssign"

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        """Identify replacement Operations and create Mutants."""
        self.mutants.append(self.create_assign_mutant(node))

        if type(node.op) in self.type_map:
            mut_types = self.type_map[type(node.op)]
            self.mutants.extend([self.create_aug_assign_mutant(node, new_type) for new_type in mut_types])

    def create_assign_mutant(self, node: ast.AugAssign) -> FileMutation:
        """Create Assign to replace AugAssign."""
        return self.create_file_mutation(
            node,
            ast.unparse(ast.Assign(lineno=node.lineno, targets=[node.target], value=node.value)),
        )

    def create_aug_assign_mutant(self, node: ast.AugAssign, new_type: type) -> FileMutation:
        """Create replacement AugAssign with alternate operation."""
        return self.create_file_mutation(
            node,
            ast.unparse(ast.AugAssign(target=node.target, op=new_type(), value=node.value)),
        )



"""Command Line Interface."""

from __future__ import annotations

import sys
import traceback
from pathlib import Path

import click

from . import (
    PoodleInputError,
    PoodleNoMutantsFoundError,
    PoodleTestingFailedError,
    PoodleTrialRunError,
    __version__,
    core,
)
from .config import build_config

CONTEXT_SETTINGS = {
    "max_content_width": 120,
}


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("sources", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.option("-c", "config_file", help="Configuration File.", type=click.Path(exists=True, path_type=Path))
@click.option("-q", "quiet", help="Quiet mode: q, qq, or qqq", count=True)
@click.option("-v", "verbose", help="Verbose mode: v, vv, or vvv", count=True)
@click.option("-w", "workers", help="Maximum number of parallel workers.", type=int)
@click.option("--exclude", help="Add a glob exclude file filter. Multiple allowed.", multiple=True)
@click.option("--only", help="Glob pattern for files to mutate. Multiple allowed.", multiple=True)
@click.option("--report", help="Enable reporter by name. Multiple allowed.", multiple=True)
@click.option("--html", help="Folder name to store HTML report in.", type=click.Path(path_type=Path))
@click.option("--json", help="File to create with JSON report.", type=click.Path(path_type=Path))
@click.option("--fail_under", help="Fail if mutation score is under this value.", type=float)
@click.version_option(version=__version__)
def main(  # noqa: C901, PLR0912
    sources: tuple[Path],
    config_file: Path | None,
    quiet: int,
    verbose: int,
    workers: int | None,
    exclude: tuple[str],
    only: tuple[str],
    report: tuple[str],
    html: Path | None,
    json: Path | None,
    fail_under: float | None,
) -> None:
    """Poodle Mutation Test Tool."""
    try:
        config = build_config(
            sources, config_file, quiet, verbose, workers, exclude, only, report, html, json, fail_under
        )
    except PoodleInputError as err:
        for arg in err.args:
            click.echo(arg)
        sys.exit(4)

    try:
        core.main_process(config)
    except PoodleTestingFailedError as err:
        for arg in err.args:
            click.echo(arg)
        sys.exit(1)
    except KeyboardInterrupt:
        click.echo("Aborted due to Keyboard Interrupt!")
        sys.exit(2)
    except PoodleTrialRunError as err:
        for arg in err.args:
            click.echo(arg)
        sys.exit(3)
    except PoodleInputError as err:
        for arg in err.args:
            click.echo(arg)
        sys.exit(4)
    except PoodleNoMutantsFoundError as err:
        for arg in err.args:
            click.echo(arg)
        sys.exit(5)
    except:  # noqa: E722
        click.echo("Aborted due to Internal Error!")
        click.echo(traceback.format_exc())
        sys.exit(3)
    sys.exit(0)


# nomut: start
if __name__ == "__main__":
    main()



from pathlib import Path
from unittest import mock

import pytest

from poodle import util
from poodle.data_types import Mutant, MutantTrial, PoodleWork
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def mock_logger():
    with mock.patch("poodle.util.logger") as logger:
        yield logger


def test_logger():
    assert util.logger.name == "poodle.util"


class TestFilesList:
    @mock.patch("poodle.util.Path")
    def test_files_list_for_folder(self, mock_path, mock_logger):
        files = [
            Path("src/project/file_1.py"),
            Path("src/project/file_test_2.py"),
        ]
        mock_folder = mock.MagicMock()
        mock_folder.rglob.return_value = iter(files)
        mock_path.return_value = mock_folder
        in_folder = mock.MagicMock()

        assert (
            util.files_list_for_folder(
                folder=in_folder,
                match_glob="*.py",
                flags=4,
                filter_globs=["test_*.py"],
            )
            == files
        )

        mock_folder.rglob.assert_called_with("*.py", flags=4, exclude=["test_*.py"])
        mock_logger.debug.assert_any_call(
            "files_list_for_folder folder=%s, match_glob=%s filter_globs=%s",
            in_folder,
            "*.py",
            ["test_*.py"],
        )
        mock_logger.debug.assert_any_call(
            "files_list_for_folder results: folder=%s files=%s",
            in_folder,
            files,
        )

    @mock.patch("poodle.util.Path")
    def test_files_list_for_source_folders(self, mock_path):
        files_1 = [
            Path("file_1.py"),
            Path("test_file_1.py"),
        ]
        files_2 = [
            Path("file_test_2.py"),
            Path("file_2_test.py"),
        ]
        mock_folder = mock.MagicMock()
        mock_folder.rglob.side_effect = [iter(files_1), iter(files_2)]
        mock_path.return_value = mock_folder

        mock_folder_1 = mock.MagicMock()
        mock_folder_2 = mock.MagicMock()

        work = PoodleWork(
            config=PoodleConfigStub(
                source_folders=[mock_folder_1, mock_folder_2],
                file_copy_flags=0,
                file_copy_filters=["test_*.py", "*_test.py"],
            )
        )

        assert util.files_list_for_source_folders(work) == {
            mock_folder_1: files_1,
            mock_folder_2: files_2,
        }

        mock_path.assert_has_calls(
            [
                mock.call(mock_folder_1),
                mock.call().rglob("*", flags=0, exclude=["test_*.py", "*_test.py"]),
                mock.call(mock_folder_2),
                mock.call().rglob("*", flags=0, exclude=["test_*.py", "*_test.py"]),
            ]
        )
        mock_folder.rglob.assert_has_calls(
            [
                mock.call("*", flags=0, exclude=["test_*.py", "*_test.py"]),
                mock.call("*", flags=0, exclude=["test_*.py", "*_test.py"]),
            ]
        )

    @mock.patch("poodle.util.files_list_for_source_folders")
    @mock.patch("poodle.util.ZipFile")
    def test_create_temp_zips(self, ZipFile, files_list_for_source_folders, mock_logger):  # noqa: N803
        work_folder = mock.MagicMock()
        work_folder.__truediv__.side_effect = lambda x: Path(x)
        files_list_for_source_folders.return_value = {
            Path("example_1"): [Path("file_1.py")],
            Path("example_2"): [Path("file_2.py")],
        }
        work = PoodleWork(
            config=PoodleConfigStub(
                work_folder=work_folder,
            )
        )
        util.create_temp_zips(work)
        work_folder.mkdir.assert_called_with(parents=True, exist_ok=True)
        files_list_for_source_folders.assert_called_with(work)
        mock_logger.info.assert_has_calls(
            [
                mock.call("Creating zip file: %s", Path("src-1.zip")),
                mock.call("Adding file: %s", Path("file_1.py")),
                mock.call("Creating zip file: %s", Path("src-2.zip")),
                mock.call("Adding file: %s", Path("file_2.py")),
            ]
        )
        assert work.folder_zips == {
            Path("example_1"): Path("src-1.zip"),
            Path("example_2"): Path("src-2.zip"),
        }
        ZipFile.assert_any_call(Path("src-1.zip"), "w")
        ZipFile.assert_any_call(Path("src-2.zip"), "w")
        target_zip = ZipFile.return_value.__enter__.return_value
        target_zip.write.assert_has_calls(
            [
                mock.call(Path("file_1.py")),
                mock.call(Path("file_2.py")),
            ]
        )


class TestDynamicImport:
    def test_dynamic_import(self, mock_logger):
        runner = util.dynamic_import("poodle.runners.command_line.runner")
        import poodle.runners.command_line

        assert runner is poodle.runners.command_line.runner

        mock_logger.debug.assert_called_with("Import object: %s", "poodle.runners.command_line.runner")


class TestPrint:
    def test_pprint_str(self):
        assert util.pprint_str({"key1": "value1", "key2": "value2"}) == "{'key1': 'value1', 'key2': 'value2'}\n"

    @mock.patch("poodle.util.StringIO")
    @mock.patch("poodle.util.pprint")
    def test_pprint_str_mock(self, pprint, StringIO):  # noqa: N803
        obj = {"key1": "value1", "key2": "value2"}
        util.pprint_str(obj)
        out = StringIO.return_value
        pprint.assert_called_with(obj, stream=out, width=150)


class TestCalcTimeout:
    def test_calc_timeout(self):
        config = PoodleConfigStub(min_timeout=10, timeout_multiplier=10)
        clean_run_results = {
            "folder": MutantTrial(mutant=None, result=None, duration=2.01),
        }
        assert round(util.calc_timeout(config, clean_run_results), 1) == 20.1

    def test_calc_timeout_mult(self):
        config = PoodleConfigStub(min_timeout=10, timeout_multiplier=5)
        clean_run_results = {
            "folder": MutantTrial(mutant=None, result=None, duration=2.01),
        }
        assert round(util.calc_timeout(config, clean_run_results), 2) == round(10.05, 2)

    def test_calc_timeout_min(self):
        config = PoodleConfigStub(min_timeout=10, timeout_multiplier=10)
        clean_run_results = {
            "folder": MutantTrial(mutant=None, result=None, duration=0.1),
        }
        assert round(util.calc_timeout(config, clean_run_results), 1) == 10.0

    def test_calc_timeout_min_20(self):
        config = PoodleConfigStub(min_timeout=20, timeout_multiplier=10)
        clean_run_results = {
            "folder": MutantTrial(mutant=None, result=None, duration=0.1),
        }
        assert round(util.calc_timeout(config, clean_run_results), 1) == 20.0


class TestMutateLines:
    def test_mutate_lines(self):
        mutant = Mutant(
            mutator_name="Example",
            lineno=2,
            col_offset=2,
            end_lineno=2,
            end_col_offset=13,
            text="Goodbye",
            source_folder=mock.MagicMock(),
            source_file=mock.MagicMock(),
        )
        file_lines = [
            "1 The quick brown fox jumps over the lazy dog",
            "2 Hello World!",
            "3 Poodles are the best",
        ]
        assert util.mutate_lines(mutant, file_lines) == [
            "1 The quick brown fox jumps over the lazy dog",
            "2 Goodbye!",
            "3 Poodles are the best",
        ]

    def test_mutate_lines_multi(self):
        mutant = Mutant(
            mutator_name="Example",
            lineno=2,
            col_offset=2,
            end_lineno=4,
            end_col_offset=3,
            text="Goodbye, T",
            source_folder=mock.MagicMock(),
            source_file=mock.MagicMock(),
        )
        file_lines = [
            "1 The quick brown fox jumps over the lazy dog",
            "2 Hello World!",
            "3 Poodles are the best",
            "4 Two are better than one",
        ]
        assert util.mutate_lines(mutant, file_lines) == [
            "1 The quick brown fox jumps over the lazy dog",
            "2 Goodbye, Two are better than one",
        ]


class TestUnifiedDiff:
    def test_create_unified_diff(self):
        mutant = Mutant(
            mutator_name="Example",
            lineno=2,
            col_offset=2,
            end_lineno=2,
            end_col_offset=13,
            text="Goodbye",
            source_folder=mock.MagicMock(),
            source_file=mock.MagicMock(),
        )
        file_lines = [
            "1 The quick brown fox jumps over the lazy dog\n",
            "2 Hello World!\n",
            "3 Poodles are the best\n",
        ]
        mutant.source_file.read_text.return_value.splitlines.return_value = file_lines
        mutant.source_file.__str__.return_value = "src/example.py"
        diff_str = (
            "--- src/example.py\n"
            "+++ [Mutant] src/example.py:2\n"
            "@@ -1,3 +1,3 @@\n"
            " 1 The quick brown fox jumps over the lazy dog\n"
            "-2 Hello World!\n"
            "+2 Goodbye!\n"
            " 3 Poodles are the best\n"
        )
        assert util.create_unified_diff(mutant) == diff_str

    def test_create_unified_diff_no_file(self):
        mutant = Mutant(
            mutator_name="Example",
            lineno=2,
            col_offset=2,
            end_lineno=2,
            end_col_offset=13,
            text="Goodbye",
            source_folder=mock.MagicMock(),
            source_file=None,
        )
        assert util.create_unified_diff(mutant) is None



from __future__ import annotations

import importlib
import re
from pathlib import Path
from unittest import mock

import pytest
from click.testing import CliRunner

from poodle import PoodleInputError, PoodleNoMutantsFoundError, PoodleTestingFailedError, PoodleTrialRunError, cli


@pytest.fixture(autouse=True)
def _setup():
    importlib.reload(cli)


@pytest.fixture()
def main_process():
    with mock.patch("poodle.cli.core.main_process") as main_process:
        yield main_process


@pytest.fixture()
def build_config():
    with mock.patch("poodle.cli.build_config") as build_config:
        yield build_config


@pytest.fixture()
def echo():
    with mock.patch("poodle.cli.click.echo") as echo:
        yield echo


@pytest.fixture()
def runner():
    return CliRunner()


class TestCliHelp:
    def test_cli_help(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert result.output.find("Usage: main [OPTIONS] [SOURCES]...") is not None
        assert result.output.find("Poodle Mutation Test Tool.") is not None
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_cli_help_max_content_width(self, runner: CliRunner):
        with mock.patch("poodle.cli.click.command") as command:
            importlib.reload(cli)
            runner.invoke(cli.main, ["--help"])
        command.assert_called_with(context_settings={"max_content_width": 120})

    def test_cli_help_config_file(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-c PATH\s+Configuration File\..*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_quiet(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-q\s+Quiet mode: q, qq, or qqq.*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_verbose(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert re.match(r".*-v\s+Verbose mode: v, vv, or vvv.*", result.output, flags=re.DOTALL) is not None

    def test_cli_help_workers(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*-w INTEGER\s+Maximum number of parallel workers\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_exclude(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--exclude TEXT\s+Add a glob exclude file filter\. Multiple allowed\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_only(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--only TEXT\s+Glob pattern for files to mutate\. Multiple allowed\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_report(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--report TEXT\s+Enable reporter by name. Multiple allowed\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_html(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--html PATH\s+Folder name to store HTML report in\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_json(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--json PATH\s+File to create with JSON report\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )

    def test_cli_help_fail_under(self, runner: CliRunner):
        result = runner.invoke(cli.main, ["--help"])
        assert result.exit_code == 0
        assert (
            re.match(
                r".*--fail_under FLOAT\s+Fail if mutation score is under this value\..*",
                result.output,
                flags=re.DOTALL,
            )
            is not None
        )


class TestInputs:
    def assert_build_config_called_with(
        self,
        build_config: mock.MagicMock,
        sources: tuple[Path] = (),  # type: ignore [assignment]
        config_file: Path | None = None,
        quiet: int = 0,
        verbose: int = 0,
        workers: int | None = None,
        exclude: tuple[str] = (),  # type: ignore [assignment]
        only: tuple[str] = (),  # type: ignore [assignment]
        report: tuple[str] = (),  # type: ignore [assignment]
        html: Path | None = None,
        json: Path | None = None,
        fail_under: float | None = None,
    ):
        build_config.assert_called_with(
            sources,
            config_file,
            quiet,
            verbose,
            workers,
            exclude,
            only,
            report,
            html,
            json,
            fail_under,
        )

    def test_cli(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config)
        main_process.assert_called_with(build_config.return_value)

    def test_main_sources(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["src", "tests"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, sources=(Path("src"), Path("tests")))  # type: ignore [arg-type]
        main_process.assert_called_with(build_config.return_value)

    def test_main_sources_not_exist(
        self,
        main_process: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        result = runner.invoke(cli.main, ["not_exist_path"])
        assert result.exit_code > 0
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_main_config_file(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-c", "pyproject.toml"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, config_file=Path("pyproject.toml"))
        main_process.assert_called_with(build_config.return_value)

    def test_main_config_file_not_exists(
        self,
        main_process: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        result = runner.invoke(cli.main, ["-c", "not_exist_path"])
        assert result.exit_code > 0
        build_config.assert_not_called()
        main_process.assert_not_called()

    def test_main_quiet(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-q"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, quiet=1)
        main_process.assert_called_with(build_config.return_value)

    def test_main_quiet2(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-qq"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, quiet=2)
        main_process.assert_called_with(build_config.return_value)

    def test_main_verbose(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-v"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, verbose=1)
        main_process.assert_called_with(build_config.return_value)

    def test_main_verbose2(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-vv"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, verbose=2)
        main_process.assert_called_with(build_config.return_value)

    def test_main_workers(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["-w", "4"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, workers=4)
        main_process.assert_called_with(build_config.return_value)

    def test_main_exclude(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--exclude", "test_.*"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, exclude=("test_.*",))
        main_process.assert_called_with(build_config.return_value)

    def test_main_only(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--only", "test_.*"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, only=("test_.*",))
        main_process.assert_called_with(build_config.return_value)

    def test_main_report(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--report", "json"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, report=("json",))
        main_process.assert_called_with(build_config.return_value)

    def test_main_html(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--html", "html_report"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, html=Path("html_report"))
        main_process.assert_called_with(build_config.return_value)

    def test_main_json(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--json", "summary.json"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, json=Path("summary.json"))
        main_process.assert_called_with(build_config.return_value)

    def test_main_fail_under(self, main_process: mock.MagicMock, build_config: mock.MagicMock, runner: CliRunner):
        result = runner.invoke(cli.main, ["--fail_under", "80"])
        assert result.exit_code == 0
        self.assert_build_config_called_with(build_config, fail_under=80)
        main_process.assert_called_with(build_config.return_value)


class TestErrors:
    def test_main_build_config_input_error(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        build_config.side_effect = PoodleInputError("bad input", "input error")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 4
        build_config.assert_called()
        echo.assert_has_calls(
            [
                mock.call("bad input"),
                mock.call("input error"),
            ]
        )
        main_process.assert_not_called()

    def test_main_process_testing_failed(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = PoodleTestingFailedError("testing failed", "error message")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 1
        build_config.assert_called()
        echo.assert_has_calls(
            [
                mock.call("testing failed"),
                mock.call("error message"),
            ]
        )
        main_process.assert_called_with(build_config.return_value)

    def test_main_process_keyboard_interrupt(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = KeyboardInterrupt()
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 2
        build_config.assert_called()
        echo.assert_called_once_with("Aborted due to Keyboard Interrupt!")
        main_process.assert_called_with(build_config.return_value)

    def test_main_process_trial_run_error(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = PoodleTrialRunError("testing failed", "error message")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 3
        build_config.assert_called()
        echo.assert_has_calls(
            [
                mock.call("testing failed"),
                mock.call("error message"),
            ]
        )
        main_process.assert_called_with(build_config.return_value)

    def test_main_process_input_error(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = PoodleInputError("testing failed", "error message")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 4
        build_config.assert_called()
        echo.assert_has_calls(
            [
                mock.call("testing failed"),
                mock.call("error message"),
            ]
        )
        main_process.assert_called_with(build_config.return_value)

    def test_main_process_no_mutants_error(
        self,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = PoodleNoMutantsFoundError("testing failed", "error message")
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 5
        build_config.assert_called()
        echo.assert_has_calls(
            [
                mock.call("testing failed"),
                mock.call("error message"),
            ]
        )
        main_process.assert_called_with(build_config.return_value)

    @mock.patch("poodle.cli.traceback")
    def test_main_process_other_error(
        self,
        traceback: mock.MagicMock,
        main_process: mock.MagicMock,
        echo: mock.MagicMock,
        build_config: mock.MagicMock,
        runner: CliRunner,
    ):
        main_process.side_effect = Exception()
        result = runner.invoke(cli.main, [])
        assert result.exit_code == 3
        build_config.assert_called()
        echo.assert_any_call("Aborted due to Internal Error!")
        echo.assert_any_call(traceback.format_exc.return_value)
        main_process.assert_called_with(build_config.return_value)



"""Create Mutants."""

from __future__ import annotations

import ast
import logging
import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable

from . import PoodleInputError
from .data_types import FileMutation, Mutant, Mutator, PoodleWork
from .mutators import (
    AugAssignMutator,
    BinaryOperationMutator,
    ComparisonMutator,
    DecoratorMutator,
    DictArrayCallMutator,
    FunctionCallMutator,
    KeywordMutator,
    LambdaReturnMutator,
    NumberMutator,
    ReturnMutator,
    StringMutator,
    UnaryOperationMutator,
)
from .util import dynamic_import, files_list_for_folder

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)

builtin_mutators = {
    BinaryOperationMutator.mutator_name: BinaryOperationMutator,
    AugAssignMutator.mutator_name: AugAssignMutator,
    UnaryOperationMutator.mutator_name: UnaryOperationMutator,
    ComparisonMutator.mutator_name: ComparisonMutator,
    KeywordMutator.mutator_name: KeywordMutator,
    NumberMutator.mutator_name: NumberMutator,
    StringMutator.mutator_name: StringMutator,
    FunctionCallMutator.mutator_name: FunctionCallMutator,
    DictArrayCallMutator.mutator_name: DictArrayCallMutator,
    LambdaReturnMutator.mutator_name: LambdaReturnMutator,
    ReturnMutator.mutator_name: ReturnMutator,
    DecoratorMutator.mutator_name: DecoratorMutator,
}


def initialize_mutators(work: PoodleWork) -> list[Callable | Mutator]:
    """Initialize all mutators from standard list and from config options."""
    mutators: list[Any] = [
        mutator for name, mutator in builtin_mutators.items() if name not in work.config.skip_mutators
    ]
    mutators.extend(work.config.add_mutators)

    return [initialize_mutator(work, mut_def) for mut_def in mutators]


def initialize_mutator(work: PoodleWork, mutator_def: Any) -> Callable | Mutator:  # noqa: ANN401
    """Import and initialize a Mutator.

    mutator_def may be string of object to import, Callable, Mutator subclass or Mutator subclass instance.
    """
    logger.debug(mutator_def)

    if isinstance(mutator_def, str):
        try:
            mutator_def = dynamic_import(mutator_def)
        except Exception as ex:  # noqa: BLE001
            msg = f"Import failed for mutator '{mutator_def}'"
            raise PoodleInputError(msg) from ex

    if isinstance(mutator_def, type) and issubclass(mutator_def, Mutator):
        return mutator_def(config=work.config, echo=work.echo)

    if callable(mutator_def) or isinstance(mutator_def, Mutator):
        return mutator_def

    msg = (
        f"Unable to create mutator '{mutator_def}' of type={type(mutator_def)}. "
        "Expected String, Callable, Mutator subclass or Mutator subclass instance."
    )
    raise PoodleInputError(msg)


def create_mutants_for_all_mutators(work: PoodleWork) -> list[Mutant]:
    """Create consolidated, flattened list of all mutants to be tried."""
    return [
        mutant
        for folder, files in get_target_files(work).items()
        for file in files
        for mutant in create_mutants_for_file(
            work,
            folder,
            file,
        )
    ]


def get_target_files(work: PoodleWork) -> dict[Path, list[Path]]:
    """Create mapping from each source folder to all mutable files in that folder."""
    logger.debug(
        "get_target_files source_folders=%s only_files=%s file_filters=%s",
        work.config.source_folders,
        work.config.only_files,
        work.config.source_folders,
    )

    if work.config.only_files:
        out: dict[Path, list[Path]] = {}
        for folder in work.config.source_folders:
            out[folder] = []
            for glob in work.config.only_files:
                out[folder] += files_list_for_folder(
                    folder=folder,
                    match_glob=glob,
                    flags=work.config.file_flags,
                    filter_globs=[],
                )
        return out
    return {
        folder: files_list_for_folder(
            folder=folder,
            match_glob="*.py",
            flags=work.config.file_flags,
            filter_globs=work.config.file_filters,
        )
        for folder in work.config.source_folders
    }


def create_mutants_for_file(work: PoodleWork, folder: Path, file: Path) -> list[Mutant]:
    """Create all mutants for specified file.

    * Parse ast from file.
    * Pass file ast to all mutators.
    * Apply Filters.
    * Compile list of Mutants.
    """
    logger.debug("Create Mutants for file %s", file)

    parsed_ast = ast.parse(file.read_bytes(), file)
    file_lines = file.read_text("utf-8").splitlines()

    def call_mutator(mutator: Callable | Mutator) -> list[FileMutation]:
        if isinstance(mutator, Mutator):
            return mutator.create_mutations(parsed_ast=deepcopy(parsed_ast), file_lines=deepcopy(file_lines))
        return mutator(config=work.config, parsed_ast=deepcopy(parsed_ast), file_lines=deepcopy(file_lines))

    mutant_nested_list = [call_mutator(mutator) for mutator in work.mutators]
    file_mutants = [mutant for mutant_list in mutant_nested_list if mutant_list for mutant in mutant_list]

    line_filters = parse_filters(file_lines)
    file_mutants = [mut for mut in file_mutants if not is_filtered(line_filters, mut)]

    return [Mutant(source_folder=folder, source_file=file, **vars(file_mutant)) for file_mutant in file_mutants]


def parse_filters(file_lines: list[str]) -> dict[int, set[str]]:
    r"""Parse text for comments to filter mutations.

    Block all mutations:

    \# pragma: no mutate

    \# nomut

    \# nomut: start

    \# nomut: end

    Block only specific mutations:

    \# nomut: mutator,mutator
    """
    line_filters: dict[int, set[str]] = {}

    no_mut_on = False

    for lineno, line in enumerate(file_lines, start=1):
        if re.search(r"#\s*pragma:\s*no mutate[\s#$]*", line):
            add_line_filter(line_filters, lineno, "all")
        no_mut_filter: list[str] = re.findall(r"#\s*nomut:?\s*([A-Za-z0-9,\s]*)[#$]*", line)

        if no_mut_filter and no_mut_filter[0].strip().lower() in ("start", "on"):
            no_mut_on = True

        if no_mut_on:
            add_line_filter(line_filters, lineno, "all")
        else:
            for mutators in no_mut_filter:
                for mutator in mutators.split(","):
                    add_line_filter(line_filters, lineno, mutator.strip())

        if no_mut_filter and no_mut_filter[0].strip().lower() in ("end", "off"):
            no_mut_on = False

    return line_filters


def add_line_filter(line_filters: dict[int, set[str]], lineno: int, mutator: str) -> None:
    """Add filter to line_filters dict."""
    if lineno not in line_filters:
        line_filters[lineno] = set()
    line_filters[lineno].add("all" if mutator == "" else mutator.lower())


def is_filtered(line_filters: dict[int, set[str]], file_mutant: FileMutation) -> bool:
    """Identify if this mutant matches any filters."""
    for lineno in range(file_mutant.lineno, file_mutant.end_lineno + 1):
        if lineno in line_filters and (
            "all" in line_filters[lineno] or file_mutant.mutator_name.lower() in line_filters[lineno]
        ):
            return True
    return False
